# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import datetime
import os
import re
import tempfile

import yaml

from oslo_config import cfg
from oslo_config import types
from shaker.engine import utils


IMAGE_BUILDER_TEMPLATES = 'shaker/resources/image_builder_templates/'
REPORT_TEMPLATES = 'shaker/resources/report_templates/'
SCENARIOS = 'shaker/scenarios/'
SCHEMAS = 'shaker/resources/schemas/'
DEFAULT_POLLING_INTERVAL = 10


class Endpoint(types.String):

    def __call__(self, value):
        value = str(value)
        if not re.match('\S+:\d+', value):
            raise ValueError('Wrong value of server_endpoint, '
                             'expected <host>:<port>, but got: %s' % value)
        return value

    def __repr__(self):
        return "Endpoint <host:port>"


class Yaml(types.String):

    def __call__(self, value):
        value = str(value)
        try:
            value = yaml.safe_load(value)
        except Exception:
            raise ValueError('YAML value is expected, but got: %s' % value)
        return value

    def __repr__(self):
        return "YAML data"


def generate_output_name():
    file_name = "shaker_%s.json" % utils.strict(str(datetime.datetime.now()))
    return os.path.join(tempfile.gettempdir(), file_name)


COMMON_OPTS = [
    cfg.Opt('server-endpoint',
            default=utils.env('SHAKER_SERVER_ENDPOINT'),
            required=True,
            type=Endpoint(),
            help='Address for server connections (host:port), '
                 'defaults to env[SHAKER_SERVER_ENDPOINT].'),
    cfg.IntOpt('polling-interval',
               default=(utils.env('SHAKER_POLLING_INTERVAL') or
                        DEFAULT_POLLING_INTERVAL),
               help='How frequently the agent polls server, in seconds')
]

OPENSTACK_OPTS = [
    cfg.StrOpt('os-auth-url', metavar='<auth-url>',
               default=utils.env('OS_AUTH_URL'),
               sample_default='',
               help='Authentication URL, defaults to env[OS_AUTH_URL].'),
    cfg.StrOpt('os-tenant-name', metavar='<auth-tenant-name>',
               default=utils.env('OS_TENANT_NAME'),
               sample_default='',
               help='Authentication tenant name, defaults to '
                    'env[OS_TENANT_NAME].'),
    cfg.StrOpt('os-project-name', metavar='<auth-project-name>',
               default=utils.env('OS_PROJECT_NAME'),
               sample_default='',
               help='Authentication project name. This option is '
                    'mutually exclusive with --os-tenant-name. '
                    'Defaults to env[OS_PROJECT_NAME].'),
    cfg.StrOpt('os-username', metavar='<auth-username>',
               default=utils.env('OS_USERNAME'),
               sample_default='',
               help='Authentication username, defaults to env[OS_USERNAME].'),
    cfg.StrOpt('os-password', metavar='<auth-password>',
               default=utils.env('OS_PASSWORD'),
               sample_default='',
               help='Authentication password, defaults to env[OS_PASSWORD].'),
    cfg.StrOpt('os-cacert', metavar='<auth-cacert>',
               default=utils.env('OS_CACERT'),
               sample_default='',
               help='Location of CA Certificate, defaults to env[OS_CACERT].'),
    cfg.BoolOpt('os-insecure',
                default=(utils.env('OS_INSECURE') or False),
                help='When using SSL in connections to the registry server, '
                     'do not require validation via a certifying authority, '
                     'defaults to env[OS_INSECURE].'),
    cfg.StrOpt('os-region-name', metavar='<auth-region-name>',
               default=utils.env('OS_REGION_NAME') or 'RegionOne',
               help='Authentication region name, defaults to '
                    'env[OS_REGION_NAME].'),

    cfg.StrOpt('external-net',
               default=utils.env('SHAKER_EXTERNAL_NET'),
               help='Name or ID of external network, defaults to '
                    'env[SHAKER_EXTERNAL_NET]. If no value provided then '
                    'Shaker picks any of available external networks.'),

    cfg.ListOpt('dns-nameservers',
                default=['8.8.8.8', '8.8.4.4'],
                help='Comma-separated list of IPs of the DNS nameservers '
                     'for the subnets. If no value is provided defaults to '
                     ' Google Public DNS.'),

    cfg.StrOpt('image-name',
               default=utils.env('SHAKER_IMAGE') or 'shaker-image',
               help='Name of image to use. The default is created by '
                    'shaker-image-builder.'),
    cfg.StrOpt('flavor-name',
               default=utils.env('SHAKER_FLAVOR') or 'shaker-flavor',
               help='Name of image flavor. The default is created by '
                    'shaker-image-builder.'),
    cfg.BoolOpt('cleanup-on-error',
                default=(utils.env('SHAKER_CLEANUP_ON_ERROR') or True),
                help='Clean up the heat-stack upon any error occurred during '
                     'scenario execution.'),

]

SERVER_AGENT_OPTS = [
    cfg.IntOpt('agent-loss-timeout',
               default=utils.env('SHAKER_AGENT_LOSS_TIMEOUT') or 60,
               help='Timeout to treat agent as lost in seconds, '
                    'defaults to env[SHAKER_AGENT_LOSS_TIMEOUT]'),
    cfg.IntOpt('agent-join-timeout',
               default=utils.env('SHAKER_AGENT_JOIN_TIMEOUT') or 600,
               help='Timeout to treat agent as join failed in seconds, '
                    'defaults to env[SHAKER_AGENT_JOIN_TIMEOUT] (time '
                    'between stack deployment and start of scenario '
                    'execution).'),
]

SCENARIO_OPTS = [
    cfg.ListOpt('scenario',
                default=utils.env('SHAKER_SCENARIO'),
                required=True,
                help=utils.make_help_options(
                    'Comma-separated list of scenarios to play. Each entity '
                    'can be a file name or one of aliases: '
                    '%s. Defaults to env[SHAKER_SCENARIO].', SCENARIOS,
                    type_filter=lambda x: x.endswith('.yaml'))),
    cfg.Opt('matrix',
            default=utils.env('SHAKER_MATRIX'),
            type=Yaml(),
            help='Set the matrix of parameters for the scenario. The value '
                 'is specified in YAML format. E.g. to override the scenario '
                 'duration one may provide: "{time: 10}", or to override list '
                 'of hosts: "{host:[ping.online.net, iperf.eenet.ee]}". When '
                 'several parameters are overridden all combinations are '
                 'tested'),
    cfg.StrOpt('output',
               default=utils.env('SHAKER_OUTPUT') or generate_output_name(),
               sample_default='',
               help='File for output in JSON format, '
                    'defaults to env[SHAKER_OUTPUT]. If it is empty, then '
                    'output will be saved to '
                    '/tmp/shaker_<time_now>.json'),
    cfg.StrOpt('artifacts-dir', default=utils.env('SHAKER_ARTIFACTS_DIR'),
               help='If specified, directs Shaker to store there all its '
                    'artifacts (output, report, subunit and book). '
                    'Defaults to env[SHAKER_ARTIFACTS_DIR].'),
    cfg.BoolOpt('no-report-on-error',
                deprecated_for_removal=True,
                default=(utils.env('SHAKER_NO_REPORT_ON_ERROR') or False),
                help='Do not generate report for failed scenarios'),
]

SERVER_OPTS = SCENARIO_OPTS + SERVER_AGENT_OPTS

REPORT_OPTS = [
    cfg.StrOpt('report-template',
               default=(utils.env('SHAKER_REPORT_TEMPLATE') or 'interactive'),
               help=utils.make_help_options(
                   'Template for report. Can be a file name or one of '
                   'aliases: %s. Defaults to "interactive".',
                   REPORT_TEMPLATES)),
    cfg.StrOpt('report',
               default=utils.env('SHAKER_REPORT'),
               help='Report file name, defaults to env[SHAKER_REPORT]. '),
    cfg.StrOpt('subunit',
               default=utils.env('SHAKER_SUBUNIT'),
               help='Subunit stream file name, defaults to '
                    'env[SHAKER_SUBUNIT].'),
    cfg.StrOpt('book',
               default=utils.env('SHAKER_BOOK'),
               help='Generate report in ReST format and store it into the '
                    'specified folder, defaults to env[SHAKER_BOOK]. '),
]

INPUT_OPTS = [
    cfg.ListOpt('input',
                default=utils.env('SHAKER_INPUT'),
                required=True,
                help='File or list of files to read test results from, '
                     'defaults to env[SHAKER_INPUT].'),
]


AGENT_OPTS = [
    cfg.StrOpt('agent-id',
               default=utils.env('SHAKER_AGENT_ID'),
               help='Agent unique id, defaults to MAC of primary interface.'),
]

IMAGE_BUILDER_OPTS = [
    cfg.StrOpt('image-builder-template',
               default=(utils.env('SHAKER_IMAGE_BUILDER_TEMPLATE') or
                        'ubuntu'),
               help=utils.make_help_options(
                   'Heat template containing receipt of building the image. '
                   'Can be a file name or one of aliases: %s. '
                   'Defaults to "ubuntu".', IMAGE_BUILDER_TEMPLATES)),
    cfg.IntOpt('flavor-ram',
               default=utils.env('SHAKER_FLAVOR_RAM') or 512,
               help='Shaker image RAM size in MB, defaults to '
                    'env[SHAKER_FLAVOR_RAM]'),
    cfg.IntOpt('flavor-vcpus',
               default=utils.env('SHAKER_FLAVOR_VCPUS') or 1,
               help='Number of cores to allocate for Shaker image, '
                    'defaults to env[SHAKER_FLAVOR_VCPUS]'),
    cfg.IntOpt('flavor-disk',
               default=utils.env('SHAKER_FLAVOR_DISK') or 3,
               help='Shaker image disk size in GB, defaults to '
                    'env[SHAKER_FLAVOR_DISK]'),
    cfg.StrOpt('image-builder-mode',
               default=utils.env('SHAKER_IMAGE_BUILDER_MODE'),
               choices=['heat', 'dib'],
               help='Image building mode: "heat" - using Heat template '
                    '(requires Glance v1 for base image upload); '
                    '"dib" - using diskimage-builder elements '
                    '(requires qemu-utils and debootstrap). If not set, '
                    'switches to "dib" if Glance v1 is not available. '
                    'Can be specified as env[SHAKER_IMAGE_BUILDER_MODE]'),
]

CLEANUP_OPTS = [
    cfg.BoolOpt('cleanup',
                default=(utils.env('SHAKER_CLEANUP') or True),
                help='Cleanup the image and the flavor.'),
]


def list_opts():
    all_opts = (COMMON_OPTS + OPENSTACK_OPTS + SERVER_OPTS + REPORT_OPTS +
                INPUT_OPTS + AGENT_OPTS + IMAGE_BUILDER_OPTS + CLEANUP_OPTS)
    yield (None, copy.deepcopy(all_opts))



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing

from oslo_log import log as logging
import zmq

from shaker.agent import agent
from shaker.engine import utils


LOG = logging.getLogger(__name__)

HEARTBEAT_AGENT = '__heartbeat'


class MessageQueue(object):
    def __init__(self, endpoint):
        _, port = utils.split_address(endpoint)

        context = zmq.Context()
        self.socket = context.socket(zmq.REP)
        self.socket.bind("tcp://*:%s" % port)
        LOG.info('Listening on *:%s', port)

        heartbeat = multiprocessing.Process(
            target=agent.work,
            kwargs=dict(agent_id=HEARTBEAT_AGENT, endpoint=endpoint,
                        ignore_sigint=True))
        heartbeat.daemon = True
        heartbeat.start()

    def __iter__(self):
        try:
            while True:
                #  Wait for next request from client
                message = self.socket.recv_json()
                LOG.debug('Received request: %s', message)

                def reply_handler(reply_message):
                    self.socket.send_json(reply_message)
                    LOG.debug('Sent reply: %s', reply_message)

                try:
                    yield message, reply_handler
                except GeneratorExit:
                    break

        except BaseException as e:
            if isinstance(e, KeyboardInterrupt):  # SIGINT is ok
                LOG.info('Process is interrupted')
            else:
                LOG.exception(e)
                raise



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import copy
import itertools
import mock
import testtools

from shaker.engine import deploy
from shaker.openstack.clients import nova

ZONE = 'zone'


def nodes_helper(*nodes):
    return [dict(host=n, zone=ZONE) for n in nodes]


class TestDeploy(testtools.TestCase):

    def test_generate_agents_alone_single_room(self):
        unique = 'UU1D'
        expected = {
            'UU1D_agent_0': {
                'id': 'UU1D_agent_0',
                'mode': 'alone',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno'},
            'UU1D_agent_1': {
                'id': 'UU1D_agent_1',
                'mode': 'alone',
                'availability_zone': '%s:dos' % ZONE,
                'zone': ZONE,
                'node': 'dos'},
        }
        accommodation = deploy.normalize_accommodation(['single_room'])
        actual = deploy.generate_agents(nodes_helper('uno', 'dos'),
                                        accommodation,
                                        unique)
        self.assertEqual(expected, actual)

    def test_generate_agents_pair_single_room(self):
        unique = 'UU1D'
        expected = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'mode': 'master',
                'availability_zone': '%s:uno' % ZONE,
                'node': 'uno',
                'zone': ZONE,
                'slave_id': 'UU1D_slave_0'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'mode': 'slave',
                'availability_zone': '%s:dos' % ZONE,
                'zone': ZONE,
                'node': 'dos'},
        }
        accommodation = deploy.normalize_accommodation(['pair', 'single_room'])
        actual = deploy.generate_agents(nodes_helper('uno', 'dos', 'tre'),
                                        accommodation,
                                        unique)
        self.assertEqual(expected, actual)

    def test_generate_agents_pair_single_room_not_enough(self):
        unique = 'UU1D'
        accommodation = deploy.normalize_accommodation(['pair', 'single_room'])
        self.assertRaises(deploy.DeploymentException, deploy.generate_agents,
                          ['uno'], accommodation, unique)

    def test_generate_agents_pair_double_room(self):
        unique = 'UU1D'
        expected = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'mode': 'master',
                'availability_zone': '%s:uno' % ZONE,
                'node': 'uno',
                'zone': ZONE,
                'slave_id': 'UU1D_slave_0'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'mode': 'slave',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno'},
            'UU1D_master_1': {
                'id': 'UU1D_master_1',
                'mode': 'master',
                'availability_zone': '%s:dos' % ZONE,
                'node': 'dos',
                'zone': ZONE,
                'slave_id': 'UU1D_slave_1'},
            'UU1D_slave_1': {
                'id': 'UU1D_slave_1',
                'master_id': 'UU1D_master_1',
                'mode': 'slave',
                'availability_zone': '%s:dos' % ZONE,
                'zone': ZONE,
                'node': 'dos'},
            'UU1D_master_2': {
                'id': 'UU1D_master_2',
                'mode': 'master',
                'availability_zone': '%s:tre' % ZONE,
                'node': 'tre',
                'zone': ZONE,
                'slave_id': 'UU1D_slave_2'},
            'UU1D_slave_2': {
                'id': 'UU1D_slave_2',
                'master_id': 'UU1D_master_2',
                'mode': 'slave',
                'availability_zone': '%s:tre' % ZONE,
                'zone': ZONE,
                'node': 'tre'},
        }
        accommodation = deploy.normalize_accommodation(['pair', 'double_room'])
        actual = deploy.generate_agents(nodes_helper('uno', 'dos', 'tre'),
                                        accommodation,
                                        unique)
        self.assertEqual(expected, actual)

    def test_generate_agents_pair_mixed_room(self):
        unique = 'UU1D'
        expected = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'mode': 'master',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno',
                'slave_id': 'UU1D_slave_0'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'mode': 'slave',
                'availability_zone': '%s:dos' % ZONE,
                'zone': ZONE,
                'node': 'dos'},
            'UU1D_master_1': {
                'id': 'UU1D_master_1',
                'mode': 'master',
                'availability_zone': '%s:dos' % ZONE,
                'zone': ZONE,
                'node': 'dos',
                'slave_id': 'UU1D_slave_1'},
            'UU1D_slave_1': {
                'id': 'UU1D_slave_1',
                'master_id': 'UU1D_master_1',
                'mode': 'slave',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno'},
        }
        accommodation = deploy.normalize_accommodation(['pair', 'mixed_room'])
        actual = deploy.generate_agents(nodes_helper('uno', 'dos'),
                                        accommodation,
                                        unique)
        self.assertEqual(expected, actual)

    def test_generate_agents_alone_single_room_double_density(self):
        unique = 'UU1D'
        expected = {
            'UU1D_agent_0': {
                'id': 'UU1D_agent_0',
                'mode': 'alone',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno'},
            'UU1D_agent_1': {
                'id': 'UU1D_agent_1',
                'mode': 'alone',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno'},
        }
        accommodation = deploy.normalize_accommodation(
            ['single_room', {'density': 2}])
        actual = deploy.generate_agents(nodes_helper('uno'),
                                        accommodation,
                                        unique)
        self.assertEqual(expected, actual)

    @mock.patch('random.sample')
    def test_generate_agents_alone_single_room_compute_nodes(self, mr):
        unique = 'UU1D'
        expected = {
            'UU1D_agent_0': {
                'id': 'UU1D_agent_0',
                'mode': 'alone',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno'},
            'UU1D_agent_1': {
                'id': 'UU1D_agent_1',
                'mode': 'alone',
                'availability_zone': '%s:duo' % ZONE,
                'zone': ZONE,
                'node': 'duo'},
        }
        compute_nodes = nodes_helper('uno', 'duo', 'tre')
        mr.side_effect = lambda x, n: x[:n]
        accommodation = deploy.normalize_accommodation(
            ['single_room', {'compute_nodes': 2}])
        actual = deploy.generate_agents(compute_nodes,
                                        accommodation,
                                        unique)
        self.assertEqual(expected, actual)

    @mock.patch('random.sample')
    def test_generate_agents_alone_single_room_density_compute_nodes(self, mr):
        unique = 'UU1D'
        expected = {
            'UU1D_agent_0': {
                'id': 'UU1D_agent_0',
                'mode': 'alone',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno'},
            'UU1D_agent_1': {
                'id': 'UU1D_agent_1',
                'mode': 'alone',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno'},
        }
        compute_nodes = nodes_helper('uno', 'duo', 'tre')
        mr.side_effect = lambda x, n: x[:n]
        accommodation = deploy.normalize_accommodation(
            ['single_room', {'compute_nodes': 1}, {'density': 2}])
        actual = deploy.generate_agents(compute_nodes,
                                        accommodation,
                                        unique)
        self.assertEqual(expected, actual)

    @mock.patch('random.sample')
    def test_generate_agents_pair_single_room_density_compute_nodes(self, mr):
        unique = 'UU1D'
        compute_nodes = nodes_helper('uno', 'duo', 'tre')
        mr.side_effect = lambda x, n: x[:n]
        accommodation = deploy.normalize_accommodation(
            ['pair', 'single_room', {'density': 4}, {'compute_nodes': 2}])
        actual = deploy.generate_agents(compute_nodes,
                                        accommodation,
                                        unique)
        self.assertEqual(8, len(actual))

    def test_generate_agents_zones_specified(self):
        unique = 'UU1D'
        expected = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'slave_id': 'UU1D_slave_0',
                'mode': 'master',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'mode': 'slave',
                'availability_zone': '%s:tre' % ZONE,
                'zone': ZONE,
                'node': 'tre'},
        }
        nodes = [
            {'host': 'uno', 'zone': ZONE},
            {'host': 'duo', 'zone': 'other-zone'},
            {'host': 'tre', 'zone': ZONE},
        ]
        accommodation = deploy.normalize_accommodation(
            ['pair', 'single_room', {'zones': [ZONE]}])
        actual = deploy.generate_agents(nodes,
                                        accommodation,
                                        unique)
        self.assertEqual(expected, actual)

    @mock.patch('random.sample')
    def test_generate_agents_zones_and_compute_nodes_specified(self, mr):
        mr.side_effect = lambda x, n: x[:n]
        unique = 'UU1D'
        expected = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'slave_id': 'UU1D_slave_0',
                'mode': 'master',
                'availability_zone': '%s:uno' % ZONE,
                'zone': ZONE,
                'node': 'uno'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'mode': 'slave',
                'availability_zone': '%s:tre' % ZONE,
                'zone': ZONE,
                'node': 'tre'},
        }
        nodes = [
            {'host': 'uno', 'zone': ZONE},
            {'host': 'duo', 'zone': 'other-zone'},
            {'host': 'tre', 'zone': ZONE},
            {'host': 'cuattro', 'zone': ZONE},
            {'host': 'cinco', 'zone': ZONE},
        ]
        accommodation = deploy.normalize_accommodation(
            ['pair', 'single_room', {'zones': [ZONE]}, {'compute_nodes': 2}])
        actual = deploy.generate_agents(nodes,
                                        accommodation,
                                        unique)
        self.assertEqual(expected, actual)

    def test_generate_agents_cross_zones(self):
        unique = 'UU1D'
        expected = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'slave_id': 'UU1D_slave_0',
                'mode': 'master',
                'availability_zone': 'nova:uno',
                'zone': 'nova',
                'node': 'uno'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'mode': 'slave',
                'availability_zone': 'vcenter:tre',
                'zone': 'vcenter',
                'node': 'tre'},
            'UU1D_master_1': {
                'id': 'UU1D_master_1',
                'slave_id': 'UU1D_slave_1',
                'mode': 'master',
                'availability_zone': 'nova:duo',
                'zone': 'nova',
                'node': 'duo'},
            'UU1D_slave_1': {
                'id': 'UU1D_slave_1',
                'master_id': 'UU1D_master_1',
                'mode': 'slave',
                'availability_zone': 'vcenter:cinco',
                'zone': 'vcenter',
                'node': 'cinco'},
        }
        nodes = [
            {'host': 'uno', 'zone': 'nova'},
            {'host': 'duo', 'zone': 'nova'},
            {'host': 'tre', 'zone': 'vcenter'},
            {'host': 'quattro', 'zone': 'nova'},
            {'host': 'cinco', 'zone': 'vcenter'},
        ]
        accommodation = deploy.normalize_accommodation(
            ['pair', 'single_room', {'zones': ['nova', 'vcenter']},
             'cross_az'])
        actual = deploy.generate_agents(nodes, accommodation, unique)
        self.assertEqual(expected, actual)

    def test_filter_agents_all_deployed(self):
        agents = {
            'UU1D_agent_0': {
                'id': 'UU1D_agent_0',
                'mode': 'alone',
                'node': 'uno'},
            'UU1D_agent_1': {
                'id': 'UU1D_agent_1',
                'mode': 'alone',
                'node': 'dos'},
        }
        stack_outputs = {
            'UU1D_agent_0_ip': '10.0.0.1',
            'UU1D_agent_0_instance_name': 'i-000001',
            'UU1D_agent_1_ip': '10.0.0.2',
            'UU1D_agent_1_instance_name': 'i-000002',
        }
        filtered = deploy.filter_agents(agents, stack_outputs)
        self.assertEqual(agents, filtered)

    def test_filter_agents_partial_deployed(self):
        agents = {
            'UU1D_agent_0': {
                'id': 'UU1D_agent_0',
                'mode': 'alone',
                'node': 'uno'},
            'UU1D_agent_1': {
                'id': 'UU1D_agent_1',
                'mode': 'alone',
                'node': 'dos'},
        }
        stack_outputs = {
            'UU1D_agent_0_ip': '10.0.0.1',
            'UU1D_agent_0_instance_name': 'i-000001',
        }
        expected = {'UU1D_agent_0': agents['UU1D_agent_0']}
        filtered = deploy.filter_agents(agents, stack_outputs)
        self.assertEqual(expected, filtered)

    def test_filter_agents_pair_single_room(self):
        agents = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'mode': 'master',
                'node': 'uno',
                'slave_id': 'UU1D_slave_0'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'mode': 'slave',
                'node': 'dos'},
        }
        stack_outputs = {
            'UU1D_master_0_ip': '10.0.0.1',
            'UU1D_master_0_instance_name': 'i-000001',
            'UU1D_slave_0_ip': '10.0.0.2',
            'UU1D_slave_0_instance_name': 'i-000002',
        }
        expected = {'UU1D_master_0': agents['UU1D_master_0'],
                    'UU1D_slave_0': agents['UU1D_slave_0']}

        filtered = deploy.filter_agents(agents, stack_outputs)
        self.assertEqual(expected, filtered)

    def test_filter_agents_pair_double_room_partially_deployed(self):
        agents = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'mode': 'master',
                'node': 'uno',
                'slave_id': 'UU1D_slave_0'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'mode': 'slave',
                'node': 'uno'},
            'UU1D_master_1': {
                'id': 'UU1D_master_1',
                'mode': 'master',
                'node': 'dos',
                'slave_id': 'UU1D_slave_1'},
            'UU1D_slave_1': {
                'id': 'UU1D_slave_1',
                'master_id': 'UU1D_master_1',
                'mode': 'slave',
                'node': 'dos'},
        }
        stack_outputs = {
            'UU1D_master_0_ip': '10.0.0.1',
            'UU1D_master_0_instance_name': 'i-000001',
            'UU1D_slave_0_ip': '10.0.0.2',
            'UU1D_slave_0_instance_name': 'i-000002',
            'UU1D_master_1_ip': '10.0.0.3',
            'UU1D_master_1_instance_name': 'i-000003',
            'UU1D_slave_1_instance_name': 'i-000004',
        }
        expected = {'UU1D_master_0': agents['UU1D_master_0'],
                    'UU1D_slave_0': agents['UU1D_slave_0'], }

        filtered = deploy.filter_agents(agents, stack_outputs)
        self.assertEqual(expected, filtered)

    def test_filter_agents_pair_single_room_with_overrides(self):
        agents = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'mode': 'master',
                'node': 'uno',
                'slave_id': 'UU1D_slave_0'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'mode': 'slave',
                'node': 'dos'},
        }
        ips = {
            'UU1D_master_0': '10.0.0.2',
            'UU1D_slave_0': '10.0.0.4',
        }
        stack_outputs = {}
        expected = {'UU1D_master_0': agents['UU1D_master_0'],
                    'UU1D_slave_0': agents['UU1D_slave_0']}

        def override(agent):
            return dict(ip=ips[agent['id']])

        filtered = deploy.filter_agents(agents, stack_outputs,
                                        override=override)
        self.assertEqual(expected, filtered)
        self.assertEqual(filtered['UU1D_master_0']['ip'], ips['UU1D_master_0'])

    def test_prepare_for_cross_az(self):
        source = [
            dict(host='uno', zone='nova'),
            dict(host='duo', zone='nova'),
            dict(host='tre', zone='vcenter'),
        ]
        expected = [
            dict(host='uno', zone='nova'),
            dict(host='tre', zone='vcenter'),
        ]
        observed = deploy.prepare_for_cross_az(source, ['nova', 'vcenter'])
        self.assertEqual(expected, observed)

    # Deployment class unit tests

    def test_deploy_local(self):
        deployment = deploy.Deployment()

        expected = {
            'local': {'id': 'local', 'mode': 'alone', 'node': 'localhost'}
        }
        agents = deployment.deploy({})

        self.assertEqual(expected, agents)

    def test_deploy_static(self):
        deployment = deploy.Deployment()

        expected = {
            'agent': {'id': 'agent', 'mode': 'alone'}
        }
        agents = deployment.deploy({'agents':
                                    [{'id': 'agent', 'mode': 'alone'}]})

        self.assertEqual(expected, agents)

    def test_deploy_template_error_when_non_initialized(self):
        deployment = deploy.Deployment()

        self.assertRaises(deploy.DeploymentException,
                          deployment.deploy, {'template': 'foo'})

    @mock.patch('shaker.openstack.clients.nova.get_available_compute_nodes')
    def test_get_compute_nodes_non_admin(self, nova_nodes_mock):
        deployment = deploy.Deployment()
        deployment.openstack_client = mock.Mock()

        def raise_error(arg):
            raise nova.ForbiddenException('err')

        nova_nodes_mock.side_effect = raise_error
        accommodation = {'compute_nodes': 4}
        expected = list(itertools.repeat({'host': None, 'zone': 'nova'}, 4))

        observed = deployment._get_compute_nodes(accommodation)

        self.assertEqual(expected, observed)

    @mock.patch('shaker.openstack.clients.nova.get_available_compute_nodes')
    def test_get_compute_nodes_non_admin_zones(self, nova_nodes_mock):
        deployment = deploy.Deployment()
        deployment.openstack_client = mock.Mock()

        def raise_error(arg):
            raise nova.ForbiddenException('err')

        nova_nodes_mock.side_effect = raise_error
        accommodation = {'compute_nodes': 4, 'zones': ['nova', 'nsx']}
        expected = [
            {'host': None, 'zone': 'nova'},
            {'host': None, 'zone': 'nsx'},
            {'host': None, 'zone': 'nova'},
            {'host': None, 'zone': 'nsx'},
        ]

        observed = deployment._get_compute_nodes(accommodation)

        self.assertEqual(expected, observed)

    @mock.patch('shaker.openstack.clients.nova.get_available_compute_nodes')
    def test_get_compute_nodes_non_admin_not_configured(self, nova_nodes_mock):
        deployment = deploy.Deployment()
        deployment.openstack_client = mock.Mock()

        def raise_error(arg):
            raise nova.ForbiddenException('err')

        nova_nodes_mock.side_effect = raise_error
        accommodation = {}

        self.assertRaises(deploy.DeploymentException,
                          deployment._get_compute_nodes, accommodation)

    def test_normalize_accommodation(self):
        origin = ['pair', 'single_room', {'compute_nodes': 2}]

        expected = collections.OrderedDict()
        expected['pair'] = True
        expected['single_room'] = True
        expected['compute_nodes'] = 2

        self.assertEqual(expected, deploy.normalize_accommodation(origin))

    def test_distribute_agents(self):
        agents = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'mode': 'master',
                'ip': '10.0.0.3',
                'slave_id': 'UU1D_slave_0'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'ip': '10.0.0.4',
                'mode': 'slave'},
            'UU1D_master_1': {
                'id': 'UU1D_master_1',
                'mode': 'master',
                'ip': '10.0.0.5',
                'slave_id': 'UU1D_slave_1'},
            'UU1D_slave_1': {
                'id': 'UU1D_slave_1',
                'master_id': 'UU1D_master_1',
                'ip': '10.0.0.6',
                'mode': 'slave'},
        }
        hosts = {
            'UU1D_master_0': '001',
            'UU1D_slave_0': '002',
            'UU1D_master_1': '003',
            'UU1D_slave_1': '004',
        }

        expected = copy.deepcopy(agents)
        for k, v in hosts.items():
            expected[k]['node'] = v

        observed = deploy.distribute_agents(agents, lambda x: hosts[x])

        # expected no changes
        self.assertEqual(agents, observed)

    # todo refactor code to use lists instead of dicts
    def _test_distribute_agents_collision(self):
        agents = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'mode': 'master',
                'ip': '10.0.0.3',
                'slave_id': 'UU1D_slave_0'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'ip': '10.0.0.4',
                'mode': 'slave'},
            'UU1D_master_1': {
                'id': 'UU1D_master_1',
                'mode': 'master',
                'ip': '10.0.0.5',
                'slave_id': 'UU1D_slave_1'},
            'UU1D_slave_1': {
                'id': 'UU1D_slave_1',
                'master_id': 'UU1D_master_1',
                'ip': '10.0.0.6',
                'mode': 'slave'},
        }
        hosts = {
            'UU1D_master_0': '001',
            'UU1D_slave_0': '001',  # collides with master_0
            'UU1D_master_1': '003',
            'UU1D_slave_1': '004',
        }

        expected = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'mode': 'master',
                'ip': '10.0.0.3',
                'node': '001',
                'slave_id': 'UU1D_slave_1'},
            'UU1D_slave_1': {
                'id': 'UU1D_slave_1',
                'master_id': 'UU1D_master_0',
                'ip': '10.0.0.6',
                'node': '004',
                'mode': 'slave'},
        }

        observed = deploy.distribute_agents(agents, lambda x: hosts[x])

        self.assertEqual(expected, observed)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import testtools

from shaker import lib


class TestLib(testtools.TestCase):

    @mock.patch('shaker.engine.quorum.make_quorum')
    def test_run_program(self, make_quorum_patch):
        quorum_mock = mock.MagicMock()
        make_quorum_patch.return_value = quorum_mock

        quorum_mock.execute = mock.Mock(
            return_value={'AGENT': {'status': 'ok', 'stdout': 'STDOUT', }})

        shaker = lib.Shaker('127.0.0.1:5999', ['AGENT'])
        res = shaker.run_program('AGENT', 'ls -al')

        self.assertDictContainsSubset(
            {'status': 'ok', 'stdout': 'STDOUT', 'agent': 'AGENT',
             'executor': 'shell', 'type': 'agent'}, res)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import traceback

from oslo_log import log as logging

from shaker.agent import agent as agent_process
from shaker.engine.executors import base as base_executors


LOG = logging.getLogger(__name__)

HEARTBEAT_AGENT = '__heartbeat'
CLEANER_AGENT = '__cleaner'


class BaseOperation(object):
    def get_agent_join_timeout(self):
        return 0

    def get_active_agent_ids(self):
        pass

    def get_default_reply(self, agent_id):
        return {'operation': 'none'}

    def get_reply(self, agent_id, start_at):
        return {}

    def process_reply(self, agent_id, message):
        return {'status': 'ok'}

    def process_error(self, agent_id, exception):
        return {'status': 'error', 'info': str(exception),
                'traceback': traceback.format_exc()}

    def process_failure(self, agent_id):
        return {'status': 'lost'}

    def process_interrupt(self, agent_id):
        return {'status': 'interrupted'}


class JoinOperation(BaseOperation):
    def __init__(self, agent_ids, polling_interval, agent_join_timeout):
        super(JoinOperation, self).__init__()
        self.agent_ids = agent_ids
        self.polling_interval = polling_interval
        self.agent_join_timeout = agent_join_timeout

    def get_agent_join_timeout(self):
        return self.agent_join_timeout

    def get_active_agent_ids(self):
        return set(self.agent_ids)

    def get_reply(self, agent_id, start_at):
        return dict(operation='configure',
                    polling_interval=self.polling_interval,
                    expected_duration=0)


class ExecuteOperation(BaseOperation):
    def __init__(self, executors):
        super(ExecuteOperation, self).__init__()
        self.executors = executors

    def get_active_agent_ids(self):
        return set(self.executors.keys())

    def get_reply(self, agent_id, start_at):
        reply = dict(operation='execute',
                     start_at=start_at,
                     command=self.executors[agent_id].get_command(),
                     expected_duration=(self.executors[agent_id].
                                        get_expected_duration()))
        return reply

    def process_reply(self, agent_id, message):
        try:
            reply = self.executors[agent_id].process_reply(message)
            r = super(ExecuteOperation, self).process_reply(agent_id, message)
            r.update(reply)
        except base_executors.ExecutorException as e:
            LOG.error('Error while processing reply: %s', e)
            r = super(ExecuteOperation, self).process_error(agent_id, e)
            r.update(e.record)
        except Exception as e:
            LOG.error('Error while processing reply: %s', e)
            r = super(ExecuteOperation, self).process_error(agent_id, e)

        return r

    def process_failure(self, agent_id):
        r = super(ExecuteOperation, self).process_failure(agent_id)
        r.update(self.executors[agent_id].process_failure())
        return r

    def process_interrupt(self, agent_id):
        r = super(ExecuteOperation, self).process_interrupt(agent_id)
        r.update(self.executors[agent_id].process_failure())
        return r


class CleanOperation(BaseOperation):
    def __init__(self, polling_interval):
        self.polling_interval = polling_interval

    def get_active_agent_ids(self):
        return {CLEANER_AGENT}

    def get_default_reply(self, agent_id):
        reply = super(CleanOperation, self).get_default_reply(agent_id)
        if agent_id != HEARTBEAT_AGENT:
            # send all agents sleep
            reply['start_at'] = time.time() + self.polling_interval * 4
        return reply


class Quorum(object):
    def __init__(self, message_queue, polling_interval, agent_loss_timeout,
                 agent_join_timeout):
        self.message_queue = message_queue
        self.polling_interval = polling_interval
        self.agent_loss_timeout = agent_loss_timeout
        self.agent_join_timeout = agent_join_timeout

    def __del__(self):
        LOG.info('Cleaning the quorum')
        self._run(CleanOperation(self.polling_interval))

    def _run(self, operation):
        current = operation.get_active_agent_ids()
        LOG.debug('Executing operation %s on agents: %s', operation, current)

        working = set()
        replied = set()
        result = {}

        start_at = time.time() + self.polling_interval * 2
        lives = dict((agent_id, start_at + operation.get_agent_join_timeout())
                     for agent_id in current)

        for message, reply_handler in self.message_queue:
            agent_id = message.get('agent_id')
            reply = operation.get_default_reply(agent_id)
            now = time.time()

            if agent_id in (current - replied):
                # message from a known not yet worked agent
                lives[agent_id] = (now + self.polling_interval * 2 +
                                   self.agent_loss_timeout)
                op = message.get('operation')

                if op == 'poll':
                    reply = operation.get_reply(agent_id, start_at)
                    lives[agent_id] += reply.get('expected_duration')
                    working.add(agent_id)
                    LOG.debug('Working agents: %s', working)
                elif op == 'reply':
                    if agent_id in working:
                        result[agent_id] = operation.process_reply(
                            agent_id, message)
                        replied.add(agent_id)
                        LOG.debug('Replied agents: %s', replied)

            reply_handler(reply)

            lost = set(a for a, t in lives.items() if t < now) - replied
            if lost:
                LOG.debug('Lost agents: %s', lost)

            if replied | lost >= current:
                if lost:
                    filtered = set(a for a in lost if a[0] != '_')
                    if filtered:  # do not warn about private agents
                        LOG.warning('Lost agents: %s', filtered)
                    # update result with info about lost agents
                    for agent_id in lost:
                        result[agent_id] = operation.process_failure(agent_id)

                LOG.info('Finished processing operation: %s', operation)
                break

        # treat missing agents as interrupted
        interrupted = current - set(result.keys())
        if interrupted:
            LOG.info('Interrupted agents: %s', interrupted)
            result.update(dict((a_id, operation.process_interrupt(a_id))
                          for a_id in interrupted))

        # update records with scheduling time
        for record in result.values():
            record['schedule'] = start_at

        return result

    def join(self, agent_ids):
        LOG.info('Waiting for quorum of agents: %s', agent_ids)
        return self._run(JoinOperation(agent_ids, self.polling_interval,
                                       self.agent_join_timeout))

    def execute(self, executors):
        return self._run(ExecuteOperation(executors))


class LocalQuorum(object):
    def execute(self, executors):
        operation = ExecuteOperation(executors)
        agent_ids = operation.get_active_agent_ids()
        result = {}

        for agent_id in agent_ids:
            task = operation.get_reply(agent_id, None)
            command_res = agent_process.run_command(task.get('command'))
            result[agent_id] = operation.process_reply(agent_id, command_res)

        return result


def make_quorum(agent_ids, message_queue, polling_interval,
                agent_loss_timeout, agent_join_timeout):
    quorum = Quorum(message_queue, polling_interval, agent_loss_timeout,
                    agent_join_timeout)
    result = quorum.join(set(agent_ids))

    failed = dict((agent_id, rec['status'])
                  for agent_id, rec in result.items() if rec['status'] != 'ok')
    if failed:
        raise Exception('Agents failed to join: %s' % failed)

    return quorum


def make_local_quorum():
    return LocalQuorum()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile

from oslo_config import cfg
from oslo_log import log as logging

from shaker.engine import config
from shaker.engine import image_builder
from shaker.engine import server
from shaker.engine import utils

LOG = logging.getLogger(__name__)


def main():
    utils.init_config_and_logging(
        config.COMMON_OPTS + config.OPENSTACK_OPTS + config.SERVER_OPTS +
        config.REPORT_OPTS + config.IMAGE_BUILDER_OPTS + config.CLEANUP_OPTS,
        use_stderr=True
    )

    artifacts_dir = cfg.CONF.artifacts_dir
    if not artifacts_dir:
        artifacts_dir = tempfile.mkdtemp(prefix='shaker')
        cfg.CONF.set_override('artifacts_dir', artifacts_dir)

    # image-builder
    LOG.info('Building the image')
    image_builder.build_image()

    # core
    if len(cfg.CONF.scenario) > 1:
        cfg.CONF.set_override(
            'output', utils.join_folder_prefix_ext(
                artifacts_dir, 'aggregated', 'json'))
        cfg.CONF.set_override(
            'report', utils.join_folder_prefix_ext(
                artifacts_dir, 'aggregated', 'html'))
        cfg.CONF.set_override(
            'subunit', utils.join_folder_prefix_ext(
                artifacts_dir, 'aggregated', 'subunit'))
        cfg.CONF.set_override(
            'book', utils.join_folder_prefix_ext(artifacts_dir, 'aggregated'))

    LOG.info('Executing scenario(s)')
    server.act()

    # cleanup
    LOG.info('Cleaning up')
    image_builder.cleanup()

if __name__ == "__main__":
    main()



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shlex
import shutil
import sys
import tempfile
import uuid

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging

from shaker.engine import config
from shaker.engine import utils
from shaker.openstack.clients import glance
from shaker.openstack.clients import heat
from shaker.openstack.clients import neutron
from shaker.openstack.clients import nova
from shaker.openstack.clients import openstack


LOG = logging.getLogger(__name__)


def init():
    openstack_params = utils.pack_openstack_params(cfg.CONF)
    try:
        return openstack.OpenStackClient(openstack_params)
    except Exception as e:
        LOG.error('Failed to connect to OpenStack: %s. '
                  'Please verify parameters: %s', e, openstack_params)
        exit(1)


def ensure_flavor(openstack_client, flavor_name):
    if nova.does_flavor_exist(openstack_client.nova, flavor_name):
        LOG.info('Using existing flavor: %s', flavor_name)
    else:
        try:
            nova.create_flavor(openstack_client.nova, name=flavor_name,
                               ram=cfg.CONF.flavor_ram,
                               vcpus=cfg.CONF.flavor_vcpus,
                               disk=cfg.CONF.flavor_disk)
            LOG.info('Created flavor %s', flavor_name)
        except nova.ForbiddenException:
            LOG.error('User does not have permissions to create the flavor. '
                      'Specify user with admin privileges or specify existing '
                      'flavor via --flavor-name parameter.')
            exit(1)


def build_image_with_heat(openstack_client, image_name, flavor_name,
                          dns_nameservers):
    template = None
    template_filename = cfg.CONF.image_builder_template
    try:
        am = lambda f: config.IMAGE_BUILDER_TEMPLATES + '%s.yaml' % f
        template = utils.read_file(template_filename, alias_mapper=am)
    except IOError:
        LOG.error('Error reading template file: %s. '
                  'Please verify correctness of --image-builder-template '
                  'parameter', template_filename)
        exit(1)
    external_net = (cfg.CONF.external_net or
                    neutron.choose_external_net(openstack_client.neutron))
    stack_name = 'shaker_%s' % uuid.uuid4()
    stack_parameters = {'external_net': external_net,
                        'flavor': flavor_name,
                        'dns_nameservers': dns_nameservers}
    stack_id = None
    try:
        stack_id = heat.create_stack(openstack_client.heat, stack_name,
                                     template, stack_parameters)

        outputs = heat.get_stack_outputs(openstack_client.heat, stack_id)
        LOG.debug('Stack outputs: %s', outputs)

        LOG.debug('Waiting for server to shutdown')
        server_id = outputs['server_info'].get('id')
        nova.wait_server_shutdown(openstack_client.nova, server_id)

        LOG.debug('Making snapshot')
        openstack_client.nova.servers.create_image(server_id, image_name)

        LOG.debug('Waiting for server to snapshot')
        nova.wait_server_snapshot(openstack_client.nova, server_id)

        LOG.info('Created image: %s', image_name)
    except BaseException as e:
        if isinstance(e, KeyboardInterrupt):
            LOG.info('Caught SIGINT. Terminating')
        else:
            error_msg = 'Error while building the image: %s' % e
            LOG.error(error_msg)
            LOG.exception(e)
    finally:
        if stack_id and cfg.CONF.cleanup_on_error:
            LOG.debug('Cleaning up the stack: %s', stack_id)
            openstack_client.heat.stacks.delete(stack_id)


def _log_multi_line(s):
    if s:
        for line in s.split('\n'):
            if line:
                LOG.debug(line)


def build_image_with_dib(openstack_client, image_name):
    elements = utils.resolve_relative_path('shaker/resources/image_elements')
    LOG.debug('Using DIB elements from: %s', elements)

    temp_dir = None
    old_cwd = os.getcwd()

    try:
        temp_dir = tempfile.mkdtemp()
        os.chdir(temp_dir)
        filename = os.path.join(temp_dir, 'shaker-image.qcow2')
        command = 'disk-image-create -o %s ubuntu vm shaker' % filename

        # path: local python binaries (including virtualenv) + sys.path + $PATH
        sys_path = ([sys.exec_prefix, os.path.join(sys.exec_prefix, 'bin')] +
                    sys.path + os.environ.get('PATH').split(':'))

        env = {}
        env.update(os.environ)
        env.update({'ELEMENTS_PATH': elements, 'PATH': ':'.join(sys_path)})

        command_stdout, command_stderr = None, None
        try:
            command_stdout, command_stderr = processutils.execute(
                *shlex.split(command), check_exit_code=True,
                root_helper='sudo', env_variables=env,
                loglevel=logging.INFO, log_errors=True)
        finally:
            _log_multi_line(command_stdout)
            _log_multi_line(command_stderr)

        LOG.info('diskimage-builder built an image %s of size %s',
                 filename, os.path.getsize(filename))

        LOG.debug('Creating image in Glance')
        image = openstack_client.glance.images.create(
            container_format='bare', disk_format='qcow2',
            name=image_name, visibility='public')

        LOG.debug('Uploading image contents into Glance')
        with open(filename, 'rb') as fd:
            openstack_client.glance.images.upload(image['id'], fd)

        LOG.info('Created image: %s', image_name)
    except Exception as e:
        LOG.error('Image build failed: %s', e)
    finally:
        if temp_dir:
            os.chdir(old_cwd)
            shutil.rmtree(temp_dir)


def ensure_image(openstack_client, image_name, flavor_name, dns_nameservers,
                 mode):
    if glance.get_image(openstack_client.glance, image_name):
        LOG.info('Using existing image: %s', image_name)
    else:
        has_v1 = (glance.get_supported_versions(openstack_client.glance) &
                  {'v1.0', 'v1.1'})

        if not has_v1 and mode == 'heat':
            LOG.warning('Glance v1.x is required to build image with Heat')

        if not mode:
            # auto-detect mode
            mode = 'heat' if has_v1 else 'dib'
            LOG.info('Detected build mode is "%s"', mode)

        if mode == 'heat':
            build_image_with_heat(openstack_client, image_name, flavor_name,
                                  dns_nameservers)
        else:
            build_image_with_dib(openstack_client, image_name)


def build_image():
    openstack_client = init()

    flavor_name = cfg.CONF.flavor_name
    image_name = cfg.CONF.image_name
    dns_nameservers = cfg.CONF.dns_nameservers
    mode = cfg.CONF.image_builder_mode

    ensure_flavor(openstack_client, flavor_name)

    ensure_image(openstack_client, image_name, flavor_name, dns_nameservers,
                 mode)


def cleanup():
    openstack_client = init()
    flavor_name = cfg.CONF.flavor_name
    image_name = cfg.CONF.image_name

    if not cfg.CONF.cleanup:
        LOG.info('Skip cleanup')
        return

    image = glance.get_image(openstack_client.glance, image_name)
    if image:
        openstack_client.glance.images.delete(image.id)

    flavor = nova.get_flavor(openstack_client.nova, flavor_name)
    if flavor:
        openstack_client.nova.flavors.delete(flavor.id)


def build_image_entry_point():
    utils.init_config_and_logging(
        config.OPENSTACK_OPTS + config.IMAGE_BUILDER_OPTS
    )
    build_image()


def cleanup_entry_point():
    utils.init_config_and_logging(config.OPENSTACK_OPTS + config.CLEANUP_OPTS)
    cleanup()


if __name__ == "__main__":
    build_image_entry_point()



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import os

import six
import testtools
import yaml

from shaker.engine import utils


class TestReport(testtools.TestCase):

    def _read_raw_file(self, file_name):
        if six.PY3:
            opener = functools.partial(open, encoding='utf8')
        else:
            opener = open
        with opener(file_name, 'r') as content_file:
            return content_file.read()

    def _iterate_files(self, root_path):
        for dir_data in os.walk(root_path):
            dir_path, dir_names, file_names = dir_data
            for file_name in file_names:
                if file_name.endswith('.yaml'):
                    yield os.path.join(dir_path, file_name)

    def test_yaml_valid(self):
        for file_name in self._iterate_files('shaker/scenarios'):
            cnt = self._read_raw_file(file_name)
            try:
                yaml.safe_load(cnt)
            except Exception as e:
                self.fail('File %s is invalid: %s' % (file_name, e))

    def test_scenario_schema_conformance(self):
        scenario_schema_file = 'shaker/resources/schemas/scenario.yaml'

        for file_name in self._iterate_files('shaker/scenarios/'):
            source_data = utils.read_yaml_file(file_name)
            schema_data = utils.read_yaml_file(scenario_schema_file)

            try:
                utils.validate_yaml(source_data, schema_data)
            except Exception as e:
                self.fail('Scenario %s does not conform to schema: %s' %
                          (file_name, e))



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

from shaker.engine import report
from shaker.engine import sla


class TestReport(testtools.TestCase):

    def test_verify_sla(self):
        records = {0: {'id': 0, 'type': 'agent', 'test': 'iperf_tcp',
                       'stats': {'bandwidth': {'avg': 700, 'min': 400}}},
                   1: {'id': 1, 'type': 'agent', 'test': 'iperf_udp',
                       'stats': {'bandwidth': {'avg': 1000, 'min': 800}}},
                   2: {'id': 2, 'type': 'agent', 'test': 'iperf_tcp',
                       'stats': {'bandwidth': {'avg': 850, 'min': 600}}}}

        tests = {
            'iperf_tcp': {
                'sla': [
                    '[type == "agent"] >> (stats.bandwidth.avg > 800)',
                    '[type == "agent"] >> (stats.bandwidth.min > 500)',
                ],
            },
            'iperf_udp': {
                'sla': [
                    '[type == "agent"] >> (stats.bandwidth.avg > 900)',
                ],
            }
        }

        sla_records = report.verify_sla(records, tests)

        self.assertIn(sla.SLAItem(records[0], sla.STATE_FALSE,
                                  'stats.bandwidth.avg > 800'), sla_records)
        self.assertIn(sla.SLAItem(records[0], sla.STATE_FALSE,
                                  'stats.bandwidth.min > 500'), sla_records)

        self.assertIn(sla.SLAItem(records[1], sla.STATE_TRUE,
                                  'stats.bandwidth.avg > 900'), sla_records)

        self.assertIn(sla.SLAItem(records[2], sla.STATE_TRUE,
                                  'stats.bandwidth.avg > 800'), sla_records)
        self.assertIn(sla.SLAItem(records[2], sla.STATE_TRUE,
                                  'stats.bandwidth.min > 500'), sla_records)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from oslo_config import cfg
from oslo_config import fixture as config_fixture_pkg
from oslotest import mockpatch
import testtools

from shaker.engine import config
from shaker.engine import server


class TestServer(testtools.TestCase):

    def test_extend_agents(self):
        agents_map = {
            'UU1D_master_0': {
                'id': 'UU1D_master_0',
                'mode': 'master',
                'node': 'uno',
                'slave_id': 'UU1D_slave_0'},
            'UU1D_slave_0': {
                'id': 'UU1D_slave_0',
                'master_id': 'UU1D_master_0',
                'mode': 'slave',
                'node': 'dos'},
        }
        agents = server._extend_agents(agents_map)
        self.assertDictContainsSubset(agents['UU1D_master_0']['slave'],
                                      agents['UU1D_slave_0'])
        self.assertDictContainsSubset(agents['UU1D_slave_0']['master'],
                                      agents['UU1D_master_0'])

    def test_pick_agents_full(self):
        agents = {}
        for i in range(10):
            agents[i] = {
                'id': i, 'mode': 'alone', 'node': 'uno',
            }
        picked = [set(a['id'] for a in arr)
                  for arr in server._pick_agents(agents, None)]
        self.assertEqual([set(range(10))], picked)

    def test_pick_agents_full_filter_slaves(self):
        agents = {}
        for i in range(10):
            agents['master_%s' % i] = {
                'id': 'master_%s' % i, 'mode': 'master', 'node': 'uno',
            }
            agents['slave_%s' % i] = {
                'id': 'slave_%s' % i, 'mode': 'slave', 'node': 'uno',
            }
        picked = [set(a['id'] for a in arr)
                  for arr in server._pick_agents(agents, None)]
        self.assertEqual([set('master_%s' % i for i in range(10))],
                         picked)

    def test_pick_agents_linear(self):
        agents = {}
        for i in range(10):
            agents[i] = {
                'id': i, 'mode': 'alone', 'node': 'uno',
            }
        picked = [set(a['id'] for a in arr)
                  for arr in server._pick_agents(agents, 'linear')]
        self.assertEqual([set(range(i + 1)) for i in range(0, 10)],
                         picked)

    def test_pick_agents_quadratic(self):
        agents = {}
        for i in range(10):
            agents[i] = {
                'id': i, 'mode': 'alone', 'node': 'uno',
            }
        picked = [set(a['id'] for a in arr)
                  for arr in server._pick_agents(agents,
                                                 'quadratic')]
        self.assertEqual([set(range(1)), set(range(2)),
                          set(range(5)), set(range(10))],
                         picked)


class TestServerPlayScenario(testtools.TestCase):

    scenario_file_name = 'folder/filename.yaml'
    deployment = {}
    scenario = {
        'title': 'Hamlet',
        'deployment': deployment,
        'file_name': scenario_file_name,
        'execution': {
            'tests': []
        }
    }

    def setUp(self):
        super(TestServerPlayScenario, self).setUp()

        conf = cfg.CONF
        self.addCleanup(conf.reset)
        self.config_fixture = self.useFixture(config_fixture_pkg.Config(conf))
        conf.register_opts(
            config.COMMON_OPTS + config.OPENSTACK_OPTS + config.SERVER_OPTS +
            config.REPORT_OPTS)

        self.config_fixture.config(server_endpoint='127.0.0.1:5999')
        self.config_fixture.config(scenario=self.scenario_file_name)

        self.useFixture(mockpatch.Patch('shaker.engine.quorum.make_quorum'))

    def assertContainsSimilarRecord(self, expected, output):
        has = False
        for actual in output['records'].values():
            has |= all(item in actual.items() for item in expected.items())
        s = 'Output should contain record similar to: %s' % expected
        self.assertTrue(has, msg=s)

    def assertNotContainsSimilarRecord(self, expected, output):
        has = False
        for actual in output['records'].values():
            has |= all(item in actual.items() for item in expected.items())
        s = 'Output should not contain record similar to: %s' % expected
        self.assertFalse(has, msg=s)

    @mock.patch('shaker.engine.server._under_openstack')
    @mock.patch('shaker.engine.server.execute')
    @mock.patch('shaker.engine.deploy.Deployment')
    def test_play_scenario(self, deploy_clz_mock, execute_mock, under_mock):
        deploy_obj = mock.Mock()
        deploy_clz_mock.return_value = deploy_obj
        under_mock.return_value = False

        def _execute(output, quorum, execution, agents, matrix=None):
            output['records'].update({'UUID': {'id': 'UUID', 'status': 'ok'}})

        execute_mock.side_effect = _execute

        deploy_obj.deploy.return_value = {
            'ID': {'id': 'ID', 'mode': 'alone'}
        }

        # act!
        output = server.play_scenario(mock.MagicMock, self.scenario)

        self.assertNotContainsSimilarRecord({'status': 'error'}, output)
        self.assertContainsSimilarRecord({
            'status': 'ok', 'id': 'UUID', 'scenario': self.scenario['title']},
            output)

        deploy_obj.deploy.assert_called_once_with(
            self.deployment, base_dir='folder',
            server_endpoint='127.0.0.1:5999')
        deploy_obj.cleanup.assert_called_once_with()

    @mock.patch('shaker.engine.deploy.Deployment')
    def test_play_scenario_with_openstack(self, deploy_clz_mock):
        deploy_obj = mock.Mock()
        deploy_clz_mock.return_value = deploy_obj
        self.config_fixture.config(os_username='user')
        self.config_fixture.config(os_password='password')
        self.config_fixture.config(os_tenant_name='tenant')
        self.config_fixture.config(os_auth_url='auth-url')

        deploy_obj.deploy.return_value = {
            'ID': {'id': 'ID', 'mode': 'alone'}
        }

        # act!
        output = server.play_scenario(mock.MagicMock, self.scenario)

        self.assertNotContainsSimilarRecord({'status': 'error'}, output)

        deploy_obj.deploy.assert_called_once_with(
            self.deployment, base_dir='folder',
            server_endpoint='127.0.0.1:5999')
        openstack_params = dict(
            auth=dict(username='user', password='password',
                      tenant_name='tenant', auth_url='auth-url'),
            os_region_name='RegionOne',
            os_cacert=None, os_insecure=False)
        deploy_obj.connect_to_openstack.assert_called_once_with(
            openstack_params, 'shaker-flavor', 'shaker-image', None,
            ['8.8.8.8', '8.8.4.4'])
        deploy_obj.cleanup.assert_called_once_with()

    @mock.patch('shaker.engine.server._under_openstack')
    @mock.patch('shaker.engine.deploy.Deployment')
    def test_play_scenario_no_agents(self, deploy_clz_mock, under_mock):
        deploy_obj = mock.Mock()
        deploy_clz_mock.return_value = deploy_obj
        under_mock.return_value = False

        deploy_obj.deploy.return_value = {}

        # act!
        output = server.play_scenario(mock.MagicMock, self.scenario)

        self.assertEqual(1, len(output['records']))
        self.assertContainsSimilarRecord({'status': 'error'}, output)

        deploy_obj.deploy.assert_called_once_with(
            self.deployment, base_dir='folder',
            server_endpoint='127.0.0.1:5999')
        deploy_obj.cleanup.assert_called_once_with()

    @mock.patch('shaker.engine.server._under_openstack')
    @mock.patch('shaker.engine.deploy.Deployment')
    def test_play_scenario_interrupted(self, deploy_clz_mock, under_mock):
        deploy_obj = mock.Mock()
        deploy_clz_mock.return_value = deploy_obj
        under_mock.return_value = False

        deploy_obj.deploy.side_effect = KeyboardInterrupt

        # act!
        output = server.play_scenario(mock.MagicMock, self.scenario)

        self.assertEqual(1, len(output['records']))
        self.assertContainsSimilarRecord({'status': 'interrupted'}, output)

        deploy_obj.deploy.assert_called_once_with(
            self.deployment, base_dir='folder',
            server_endpoint='127.0.0.1:5999')
        deploy_obj.cleanup.assert_called_once_with()


class TestServerExecute(testtools.TestCase):

    @mock.patch('shaker.engine.utils.make_record_id', return_value='UUID')
    def test_execute(self, mri_mock):
        output = dict(records={}, tests={})
        quorum = mock.Mock()
        quorum.execute.return_value = {'the-agent': {'status': 'ok'}}
        execution = {'tests': [{'title': 'tcp', 'class': 'iperf'}]}
        agents = {'the-agent': {'id': 'the-agent', 'mode': 'alone'}}

        server.execute(output, quorum, execution, agents)

        expected_records = {
            'UUID': {
                'agent': 'the-agent', 'concurrency': 1, 'executor': 'iperf',
                'id': 'UUID', 'node': None, 'status': 'ok', 'test': 'tcp',
                'type': 'agent'
            }
        }
        expected_tests = {
            'tcp': {'title': 'tcp', 'class': 'iperf'}
        }

        self.assertEqual(expected_records, output['records'])
        self.assertEqual(expected_tests, output['tests'])

    @mock.patch('shaker.engine.utils.make_record_id', return_value='UUID')
    def test_execute_interrupted(self, mri_mock):
        output = dict(records={}, tests={})
        quorum = mock.Mock()
        quorum.execute.return_value = {'the-agent': {'status': 'interrupted'}}
        execution = {'tests': [{'title': 'tcp', 'class': 'iperf'},
                               {'title': 'udp', 'class': 'netperf'}]}
        agents = {'the-agent': {'id': 'the-agent', 'mode': 'alone'}}

        server.execute(output, quorum, execution, agents)

        expected_records = {
            'UUID': {
                'agent': 'the-agent', 'concurrency': 1, 'executor': 'iperf',
                'id': 'UUID', 'node': None, 'status': 'interrupted',
                'test': 'tcp', 'type': 'agent'
            }
        }
        expected_tests = {
            'tcp': {'title': 'tcp', 'class': 'iperf'}
        }

        self.assertEqual(expected_records, output['records'])
        self.assertEqual(expected_tests, output['tests'])



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

from shaker.engine.executors import flent


IP = '10.0.0.10'
AGENT = {'slave': {'ip': IP}}


class TestFlentExecutor(testtools.TestCase):

    def test_get_command(self):
        executor = flent.FlentExecutor({}, AGENT)

        expected = {'data': (flent.FLENT_EXEC % 'flent -H %s -l 60 -s 1 '
                             'tcp_download') % IP,
                    'type': 'script'}
        self.assertEqual(expected, executor.get_command())

    def test_get_command_with_params(self):
        executor = flent.FlentExecutor(
            dict(method='ping', time=10, interval=0.5), AGENT)

        expected = {'data': (flent.FLENT_EXEC % 'flent -H %s -l 10 -s 0.5 '
                             'ping') % IP,
                    'type': 'script'}
        self.assertEqual(expected, executor.get_command())

    def test_get_command_static_host(self):
        executor = flent.FlentExecutor({'host': '10.0.0.20'}, {})

        expected = {'data': (flent.FLENT_EXEC % 'flent -H %s -l 60 -s 1 '
                             'tcp_download') % '10.0.0.20',
                    'type': 'script'}
        self.assertEqual(expected, executor.get_command())

    def test_get_expected_duration(self):
        executor = flent.FlentExecutor(dict(method='ping', time=10), AGENT)
        expected = 20
        self.assertEqual(expected, executor.get_expected_duration())

    def test_process_reply(self):
        executor = flent.FlentExecutor({}, AGENT)
        message = {
            'stdout': """
{
  "metadata": {
    "SERIES_META": {
      "Ping ICMP": {
        "UNITS": "ms"
      },
      "TCP upload": {
        "MEAN_VALUE": 14536.94,
        "UNITS": "Mbit/s"
      }
    },
    "TOTAL_LENGTH": 6
  },
  "results": {
    "Ping ICMP": [
      0.11,
      0.08003925186913663,
      0.08997269229670403,
      0.07008327936919093,
      0.07994293923635702
    ],
    "TCP upload": [
      null,
      9789.93,
      17333.075575393304,
      18173.561165616233,
      null
    ]
  },
  "x_values": [
    0.0,
    0.2,
    0.4,
    0.6000000000000001,
    0.8
  ]
}
"""
        }
        expected = {
            'samples': [
                [0.0, 0.11, None],
                [0.2, 0.08003925186913663, 9789.93],
                [0.4, 0.08997269229670403, 17333.075575393304],
                [0.6000000000000001, 0.07008327936919093, 18173.561165616233],
                [0.8, 0.07994293923635702, None]
            ],
            'meta': [
                ['time', 's'], ['ping_icmp', 'ms'], ['tcp_upload', 'Mbit/s'],
            ]
        }
        reply = executor.process_reply(message)
        self.assertEqual(expected['samples'], reply['samples'],
                         message='Samples data')
        self.assertEqual(expected['meta'], reply['meta'],
                         message='Metadata')



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os_client_config
from oslo_log import log as logging


LOG = logging.getLogger(__name__)


class OpenStackClient(object):
    def __init__(self, openstack_params):
        LOG.debug('Establishing connection to OpenStack')

        config = os_client_config.OpenStackConfig()
        cloud_config = config.get_one_cloud(**openstack_params)
        if openstack_params['os_insecure']:
            cloud_config.config['verify'] = False
            cloud_config.config['cacert'] = None
        self.keystone_session = cloud_config.get_session()
        self.nova = cloud_config.get_legacy_client('compute')
        self.neutron = cloud_config.get_legacy_client('network')
        self.glance = cloud_config.get_legacy_client('image')

        # heat client wants endpoint to be always set
        endpoint = cloud_config.get_session_endpoint('orchestration')
        self.heat = cloud_config.get_legacy_client('orchestration',
                                                   endpoint=endpoint)

        # Ping OpenStack
        self.keystone_session.get_token()

        LOG.info('Connection to OpenStack is initialized')



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import functools
import itertools
import logging as std_logging
import os
import random
import re
import uuid

import collections
from oslo_config import cfg
from oslo_log import log as logging
from pykwalify import core as pykwalify_core
from pykwalify import errors as pykwalify_errors
import yaml


LOG = logging.getLogger(__name__)


def env(*_vars, **kwargs):
    """Returns the first environment variable set.

    If none are non-empty, defaults to '' or keyword arg default.
    """
    for v in _vars:
        value = os.environ.get(v)
        if value:
            return value
    return kwargs.get('default', None)


def validate_required_opts(conf, opts):
    # all config parameters default to ENV values, that's why standard
    # check of required options doesn't work and needs to be done manually
    for opt in opts:
        if opt.required and not conf[opt.dest]:
            raise cfg.RequiredOptError(opt.name)


def init_config_and_logging(opts, **conf_overrides):
    conf = cfg.CONF
    conf.register_cli_opts(opts)
    conf.register_opts(opts)
    logging.register_options(conf)

    for k, v in conf_overrides.items():
        conf.set_override(k, v)

    # requests to OpenStack services should be visible at DEBUG level
    default_log_levels = [l for l in conf.default_log_levels
                          if not l.startswith('keystoneauth')]
    default_log_levels += ['pykwalify=INFO']
    logging.set_defaults(default_log_levels=default_log_levels)

    try:
        conf(project='shaker')
        validate_required_opts(conf, opts)
    except cfg.RequiredOptError as e:
        print('Error: %s' % e)
        conf.print_usage()
        exit(1)

    logging.setup(conf, 'shaker')
    LOG.info('Logging enabled')
    conf.log_opt_values(LOG, std_logging.DEBUG)


def resolve_relative_path(file_name):
    path = os.path.normpath(os.path.join(
        os.path.dirname(__import__('shaker').__file__), '../', file_name))
    if os.path.exists(path):
        return path


def read_file(file_name, base_dir='', alias_mapper=None):
    full_path = os.path.normpath(os.path.join(base_dir, file_name))

    if alias_mapper:  # interpret file_name as alias
        alias_path = resolve_relative_path(alias_mapper(file_name))
        if alias_path:
            full_path = alias_path
            LOG.info('Alias "%s" is resolved into file "%s"',
                     file_name, full_path)

    if not os.path.exists(full_path):
        # treat file_name as relative to shaker's package root
        full_path = os.path.normpath(os.path.join(
            os.path.dirname(__import__('shaker').__file__), '../', file_name))
        if not os.path.exists(full_path):
            msg = ('File %s not found by absolute nor by relative path' %
                   file_name)
            LOG.error(msg)
            raise IOError(msg)

    fd = None
    try:
        fd = open(full_path)
        return fd.read()
    except IOError as e:
        LOG.error('Error reading file: %s', e)
        raise
    finally:
        if fd:
            fd.close()


def write_file(data, file_name, base_dir=''):
    full_path = os.path.normpath(os.path.join(base_dir, file_name))
    fd = None
    try:
        fd = open(full_path, 'w')
        return fd.write(data)
    except IOError as e:
        LOG.error('Error writing file: %s', e)
        raise
    finally:
        if fd:
            fd.close()


def read_yaml_file(file_name):
    raw = read_file(file_name)
    try:
        parsed = yaml.safe_load(raw)
        return parsed
    except Exception as e:
        LOG.error('Failed to parse file %(file)s in YAML format: %(err)s',
                  dict(file=file_name, err=e))
        raise


def split_address(address):
    try:
        host, port = address.split(':')
    except ValueError:
        raise ValueError('Invalid address: %s, "host:port" expected', address)
    return host, port


def random_string(length=6):
    return ''.join(random.sample('adefikmoprstuz', length))


def make_record_id():
    return str(uuid.uuid4())


def copy_dict_kv(source):
    return dict((k, v) for k, v in source.items())


def flatten_dict(d, prefix='', sep='.'):
    res = []
    for k, v in d.items():
        path = prefix + k
        if isinstance(v, dict):
            res.extend(flatten_dict(v, path + sep))
        else:
            res.append((path, v))
    return res


def merge_dicts(src):
    res = collections.defaultdict(dict)
    for one in src:
        for k in one.keys():
            res[k].update(one[k])
    return res


def make_help_options(message, base, type_filter=None):
    path = resolve_relative_path(base)
    files = itertools.chain.from_iterable(
        [map(functools.partial(os.path.join, root), files)
         for root, dirs, files in os.walk(path)])  # list of files in a tree
    if type_filter:
        files = (f for f in files if type_filter(f))  # filtered list
    rel_files = map(functools.partial(os.path.relpath, start=path), files)
    return message % ', '.join('"%s"' % f.partition('.')[0]
                               for f in sorted(rel_files))


def algebraic_product(**kwargs):
    position_to_key = {}
    values = []
    total = 1

    for key, item in kwargs.items():
        position_to_key[len(values)] = key
        if type(item) != list:
            item = [item]  # enclose single item into the list

        values.append(item)
        total *= len(item)

    LOG.debug('Total number of permutations is: %s', total)

    for chain in itertools.product(*values):
        result = {}
        for position, key in position_to_key.items():
            result[key] = chain[position]
        yield result


def strict(s):
    return re.sub(r'[^\w\d]+', '_', re.sub(r'\(.+\)', '', s)).lower()


def validate_yaml(data, schema):
    c = pykwalify_core.Core(source_data=data, schema_data=schema)
    try:
        c.validate(raise_exception=True)
    except pykwalify_errors.SchemaError as e:
        raise Exception('File does not conform to schema: %s' % e)


def get_value_by_path(src, param):
    tokens = param.split('.')
    for token in tokens:
        if token not in src:
            return None
        src = src[token]
    return src


def set_value_by_path(dst, param, value):
    tokens = param.split('.')
    for token in tokens[:-1]:
        if token not in dst:
            dst[token] = {}
        dst = dst[token]
    dst[tokens[-1]] = value


def copy_value_by_path(src, src_param, dst, dst_param):
    v = get_value_by_path(src, src_param)
    if v:
        set_value_by_path(dst, dst_param, v)
        return True
    return False


class MisconfigurationException(Exception):
    pass


def pack_openstack_params(conf):
    if not conf.os_auth_url:
        raise MisconfigurationException(
            'OpenStack authentication endpoint is missing')

    params = dict(auth=dict(username=conf.os_username,
                            password=conf.os_password,
                            auth_url=conf.os_auth_url),
                  os_region_name=conf.os_region_name,
                  os_cacert=conf.os_cacert,
                  os_insecure=conf.os_insecure)
    if conf.os_tenant_name:
        params['auth']['tenant_name'] = conf.os_tenant_name
    if conf.os_project_name:
        params['auth']['project_name'] = conf.os_project_name
    return params


def mkdir_tree(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def join_folder_prefix_ext(folder, prefix, ext=None):
    return os.path.join(folder, '%s.%s' % (prefix, ext) if ext else prefix)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def get_image(glance_client, image_name):
    for image in glance_client.images.list():
        if image.name == image_name:
            return image
    return None


def get_supported_versions(glance_client):
    return set(version['id'] for version in glance_client.versions.list())



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging


LOG = logging.getLogger(__name__)


def choose_external_net(neutron_client):
    ext_nets = neutron_client.list_networks(
        **{'router:external': True})['networks']
    if not ext_nets:
        raise Exception('No external networks found')
    return ext_nets[0]['name']



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import functools
import random

import jinja2
from oslo_config import cfg
from oslo_log import log as logging

from shaker.engine import utils
from shaker.openstack.clients import heat
from shaker.openstack.clients import neutron
from shaker.openstack.clients import nova
from shaker.openstack.clients import openstack


LOG = logging.getLogger(__name__)


class DeploymentException(Exception):
    pass


def prepare_for_cross_az(compute_nodes, zones):
    if len(zones) != 2:
        LOG.warn('cross_az is specified, but len(zones) is not 2')
        return compute_nodes

    masters = []
    slaves = []
    for node in compute_nodes:
        if node['zone'] == zones[0]:
            masters.append(node)
        else:
            slaves.append(node)

    res = []
    for i in range(min(len(masters), len(slaves))):
        res.append(masters[i])
        res.append(slaves[i])

    return res


def generate_agents(compute_nodes, accommodation, unique):
    density = accommodation.get('density') or 1

    zones = accommodation.get('zones')
    if zones:
        compute_nodes = [c for c in compute_nodes if c['zone'] in zones]
        if 'cross_az' in accommodation:
            # sort nodes to interleave hosts from different zones
            compute_nodes = prepare_for_cross_az(compute_nodes, zones)

    if accommodation.get('compute_nodes'):
        compute_nodes = random.sample(
            compute_nodes, accommodation.get('compute_nodes'))

    cn_count = len(compute_nodes)
    iterations = cn_count * density

    if 'single_room' in accommodation and 'pair' in accommodation:
        iterations //= 2
    node_formula = lambda x: compute_nodes[x % cn_count]

    agents = {}

    for i in range(iterations):
        if 'pair' in accommodation:
            master_id = '%s_master_%s' % (unique, i)
            slave_id = '%s_slave_%s' % (unique, i)
            master = dict(id=master_id, mode='master', slave_id=slave_id)
            slave = dict(id=slave_id, mode='slave', master_id=master_id)

            if 'single_room' in accommodation:
                master_formula = lambda x: i * 2
                slave_formula = lambda x: i * 2 + 1
            elif 'double_room' in accommodation:
                master_formula = lambda x: i
                slave_formula = lambda x: i
            else:  # mixed_room
                master_formula = lambda x: i
                slave_formula = lambda x: i + 1

            m = node_formula(master_formula(i))
            master['node'], master['zone'] = m['host'], m['zone']
            s = node_formula(slave_formula(i))
            slave['node'], slave['zone'] = s['host'], s['zone']

            agents[master['id']] = master
            agents[slave['id']] = slave
        else:
            if 'single_room' in accommodation:
                agent_id = '%s_agent_%s' % (unique, i)
                agents[agent_id] = dict(id=agent_id,
                                        node=node_formula(i)['host'],
                                        zone=node_formula(i)['zone'],
                                        mode='alone')

    if not agents:
        raise DeploymentException('Not enough compute nodes %(cn)s for '
                                  'requested instance accommodation %(acc)s' %
                                  dict(cn=compute_nodes, acc=accommodation))

    # inject availability zone
    for agent in agents.values():
        az = agent['zone']
        if agent['node']:
            az += ':' + agent['node']
        agent['availability_zone'] = az

    return agents


def _get_stack_values(stack_outputs, vm_name, params):
    result = {}
    for param in params:
        o = stack_outputs.get(vm_name + '_' + param)
        if o:
            result[param] = o
    return result


def filter_agents(agents, stack_outputs, override=None):
    deployed_agents = {}

    # first pass, ignore non-deployed
    for agent in agents.values():
        stack_values = _get_stack_values(stack_outputs, agent['id'], ['ip'])

        if override:
            stack_values.update(override(agent))

        if not stack_values.get('ip'):
            LOG.info('Ignore non-deployed agent: %s', agent)
            continue

        agent.update(stack_values)

        # workaround of Nova bug 1422686
        if agent.get('mode') == 'slave' and not agent.get('ip'):
            LOG.info('IP address is missing in agent: %s', agent)
            continue

        deployed_agents[agent['id']] = agent

    # second pass, check pairs
    result = {}
    for agent in deployed_agents.values():
        if (agent.get('mode') == 'alone' or
                (agent.get('mode') == 'master' and
                 agent.get('slave_id') in deployed_agents) or
                (agent.get('mode') == 'slave' and
                 agent.get('master_id') in deployed_agents)):
            result[agent['id']] = agent

    return result


def distribute_agents(agents, get_host_fn):
    result = {}

    hosts = set()
    buckets = collections.defaultdict(list)
    for agent in agents.values():
        agent_id = agent['id']
        # we assume that server name equals to agent_id
        host_id = get_host_fn(agent_id)

        if host_id not in hosts:
            hosts.add(host_id)
            agent['node'] = host_id
            buckets[agent['mode']].append(agent)
        else:
            LOG.info('Filter out agent %s, host %s is already occupied',
                     agent_id, host_id)

    if buckets['alone']:
        result = dict((a['id'], a) for a in buckets['alone'])
    else:
        for master, slave in zip(buckets['master'], buckets['slave']):
            master['slave_id'] = slave['id']
            slave['master_id'] = master['id']

            result[master['id']] = master
            result[slave['id']] = slave

    return result


def normalize_accommodation(accommodation):
    result = {}

    for s in accommodation:
        if isinstance(s, dict):
            result.update(s)
        else:
            result[s] = True

    return result


class Deployment(object):
    def __init__(self):
        self.openstack_client = None
        self.has_stack = False
        self.privileged_mode = True

    def connect_to_openstack(self, openstack_params, flavor_name, image_name,
                             external_net, dns_nameservers):
        LOG.debug('Connecting to OpenStack')

        self.openstack_client = openstack.OpenStackClient(openstack_params)

        self.flavor_name = flavor_name
        self.image_name = image_name
        self.external_net = (external_net or
                             neutron.choose_external_net(
                                 self.openstack_client.neutron))
        self.stack_name = 'shaker_%s' % utils.random_string()
        self.dns_nameservers = dns_nameservers

    def _get_compute_nodes(self, accommodation):
        try:
            return nova.get_available_compute_nodes(self.openstack_client.nova)
        except nova.ForbiddenException:
            # user has no permissions to list compute nodes
            LOG.info('OpenStack user does not have permission to list compute '
                     'nodes - treat him as non-admin')
            self.privileged_mode = False
            count = accommodation.get('compute_nodes')
            if not count:
                raise DeploymentException(
                    'When run with non-admin user the scenario must specify '
                    'number of compute nodes to use')

            zones = accommodation.get('zones') or ['nova']
            return [dict(host=None, zone=zones[n % len(zones)])
                    for n in range(count)]

    def _deploy_from_hot(self, specification, server_endpoint, base_dir=None):
        accommodation = normalize_accommodation(
            specification.get('accommodation') or
            specification.get('vm_accommodation'))

        agents = generate_agents(self._get_compute_nodes(accommodation),
                                 accommodation, self.stack_name)

        # render template by jinja
        vars_values = {
            'agents': agents,
            'unique': self.stack_name,
        }
        heat_template = utils.read_file(specification['template'],
                                        base_dir=base_dir)
        compiled_template = jinja2.Template(heat_template)
        rendered_template = compiled_template.render(vars_values)
        LOG.debug('Rendered template: %s', rendered_template)

        # create stack by Heat
        merged_parameters = {
            'server_endpoint': server_endpoint,
            'external_net': self.external_net,
            'image': self.image_name,
            'flavor': self.flavor_name,
            'dns_nameservers': self.dns_nameservers,
        }
        merged_parameters.update(specification.get('template_parameters', {}))

        self.has_stack = True
        stack_id = heat.create_stack(
            self.openstack_client.heat, self.stack_name, rendered_template,
            merged_parameters)

        # get info about deployed objects
        outputs = heat.get_stack_outputs(self.openstack_client.heat, stack_id)
        override = self._get_override(specification.get('override'))

        agents = filter_agents(agents, outputs, override)

        if (not self.privileged_mode) and accommodation.get('density', 1) == 1:
            get_host_fn = functools.partial(nova.get_server_host_id,
                                            self.openstack_client.nova)
            agents = distribute_agents(agents, get_host_fn)

        return agents

    def _get_override(self, override_spec):
        def override_ip(agent, ip_type):
            return dict(ip=nova.get_server_ip(
                self.openstack_client.nova, agent['id'], ip_type))

        if override_spec:
            if override_spec.get('ip'):
                return functools.partial(override_ip,
                                         ip_type=override_spec.get('ip'))

    def deploy(self, deployment, base_dir=None, server_endpoint=None):
        agents = {}

        if not deployment:
            # local mode, create fake agent
            agents.update(dict(local=dict(id='local', mode='alone',
                                          node='localhost')))

        if deployment.get('template'):
            if not self.openstack_client:
                raise DeploymentException(
                    'OpenStack client is not initialized. '
                    'Template-based deployment is ignored.')
            else:
                # deploy topology specified by HOT
                agents.update(self._deploy_from_hot(
                    deployment, server_endpoint, base_dir=base_dir))

        if deployment.get('agents'):
            # agents are specified statically
            agents.update(dict((a['id'], a) for a in deployment.get('agents')))

        return agents

    def cleanup(self):
        if self.has_stack and cfg.CONF.cleanup_on_error:
            LOG.debug('Cleaning up the stack: %s', self.stack_name)
            self.openstack_client.heat.stacks.delete(self.stack_name)



# Copyright (c) 2016 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import numbers
import os
import textwrap

from oslo_log import log as logging
import pygal
from pygal import style
import six
import yaml

from shaker.engine import utils

LOG = logging.getLogger(__name__)

TABLE_FLOAT_PREC = 2


class ReSTPublisher(object):
    header_marks = ['*', '=', '-', '^', '~']

    def __init__(self, folder):
        self.folder = folder

        LOG.info('Create ReST book in: %s', folder)
        utils.mkdir_tree(folder)
        self.index = open(os.path.join(folder, 'index.rst'), 'w+')

    def __del__(self):
        self.index.close()

    def ref_label(self, text):
        self.index.write('.. _%s:\n\n' % utils.strict(text))

    def header(self, text, level=0):
        self.index.write(text)
        self.index.write('\n')
        self.index.write(self.header_marks[level] * len(text))
        self.index.write('\n\n')

    def subheader(self, text):
        self.index.write('**%s**:' % text)
        self.index.write('\n\n')

    def para(self, text):
        self.index.write(textwrap.fill(text, width=79))
        self.index.write('\n\n')

    def code(self, text):
        self.index.write('.. code-block:: yaml\n\n')
        for line in text.split('\n'):
            if line:
                self.index.write(' ' * 4 + line + '\n')
        self.index.write('\n')

    def chart_line(self, chart_id, samples, meta, x_title):
        line_chart = pygal.Line(style=style.RedBlueStyle,
                                fill=False,
                                legend_at_bottom=True,
                                include_x_axis=True,
                                x_title=x_title)

        for i in range(1, len(meta)):
            line_title = meta[i][0]
            if meta[i][1]:
                line_title += ', %s' % meta[i][1]
            kwargs = dict(secondary=True) if i == 2 else {}
            values = [samples[x][i] for x in range(len(samples))]
            line_chart.add(line_title, values, **kwargs)

        line_chart.render_to_file(os.path.join(self.folder,
                                               '%s.svg' % chart_id))
        self.index.write('.. image:: %s.*\n\n' % chart_id)

    def chart_xy(self, chart_id, chart, meta, x_title):
        xy_chart = pygal.XY(style=style.RedBlueStyle,
                            legend_at_bottom=True,
                            fill=False,
                            include_x_axis=True,
                            x_title=x_title)

        for i in range(1, len(meta)):
            line_title = meta[i][0]
            if meta[i][1]:
                line_title += ', %s' % meta[i][1]
            v = [(chart[0][j], chart[i][j]) for j in range(1, len(chart[i]))]
            kwargs = dict(secondary=True) if i == 2 else {}
            xy_chart.add(line_title, v, **kwargs)

        xy_chart.render_to_file(os.path.join(self.folder,
                                             '%s.svg' % chart_id))
        self.index.write('.. image:: %s.*\n\n' % chart_id)

    def _outline(self, widths):
        s = '  '.join('=' * w for w in widths)
        self.index.write(s)
        self.index.write('\n')

    def table(self, t):
        widths = [max(len(c), TABLE_FLOAT_PREC + 6) for c in t[0]]

        for r in t:
            for i in range(len(widths)):
                if isinstance(r[i], six.string_types):
                    widths[i] = max(widths[i], len(r[i]))

        # header
        self._outline(widths)
        self.index.write('  '.join(('{0:<{1}}'.format(t[0][i], widths[i]))
                                   for i in range(len(widths))))
        self.index.write('\n')
        self._outline(widths)

        # body
        for r in t[1:]:
            cells = []
            for i in range(len(widths)):
                c = r[i]
                if isinstance(c, numbers.Integral):
                    c = '{0:>{1}}'.format(c, widths[i])
                elif isinstance(c, numbers.Number):
                    c = '{0:>{1}.{2}f}'.format(c, widths[i], TABLE_FLOAT_PREC)
                else:
                    c = '{0:<{1}}'.format(c, widths[i])
                cells.append(c)

            self.index.write('  '.join(cells).rstrip())
            self.index.write('\n')

        # bottom border
        self._outline(widths)
        self.index.write('\n')


yamlize = functools.partial(yaml.safe_dump, indent=2, default_flow_style=False)


def filter_records(records, **kwargs):
    result = []
    for r in records:
        f = True
        for param, value in kwargs.items():
            f &= r.get(param) == value

        if f:
            result.append(r)
    return result


def write_scenario_definition(publisher, scenario):
    publisher.subheader('Scenario')
    publisher.code(yamlize(scenario))


def write_test_definition(data, publisher, test):
    publisher.subheader('Test Specification')
    publisher.code(yamlize(data['tests'][test]))


def write_sla(publisher, records, sla_records):
    table = [['Expression', 'Concurrency', 'Node', 'Result']]

    for sla_record in sla_records:
        expression = sla_record['expression']
        record_id = sla_record['record']
        state = sla_record['state']

        for record in records:
            if record['id'] == record_id:
                table.append([expression, record['concurrency'],
                              record['node'], state])
                break

    if len(table) > 1:
        publisher.subheader('SLA')
        publisher.table(table)


def write_record_stats(publisher, record):
    table = [['Metric', 'Min', 'Avg', 'Max']]
    stats = record['stats']

    for key in stats.keys():
        metric = key
        unit = stats[key].get('unit') or ''
        if unit:
            metric += ', %s' % unit

        s = stats[key]
        value_min = s.get('min') or ''
        value_avg = s.get('avg') or ''
        value_max = s.get('max') or ''

        table.append([metric, value_min, value_avg, value_max])

    if len(table) > 1:
        publisher.subheader('Stats')
        publisher.table(table)


def write_errors(publisher, records):
    bad_records = [r for r in records if r.get('status') in {'lost', 'error'}]
    if bad_records:
        publisher.subheader('Errors')
        for rec in bad_records:
            publisher.code(yamlize(rec))


def write_concurrency_block(publisher, all_records, local_records, sla):
    for record in local_records:
        concurrency = record['concurrency']
        if len(local_records) > 2:
            publisher.header('Concurrency %s' % concurrency, level=2)

        agent_records = filter_records(all_records,
                                       type='agent',
                                       scenario=record['scenario'],
                                       test=record['test'],
                                       concurrency=concurrency)

        agent_records_ok = filter_records(agent_records, status='ok')

        write_errors(publisher, agent_records)

        if len(agent_records_ok) <= 2 and len(local_records) <= 2:
            # go into details
            write_agent_block_detailed(publisher, agent_records_ok)
        else:
            # show stats only
            write_stats(publisher, agent_records_ok, 'node')

        write_sla(publisher, agent_records_ok, sla)


def write_agent_block_detailed(publisher, records):
    for record in records:
        if len(records) > 1:
            publisher.header('Agent %s' % record['agent'], level=3)

        if record.get('samples'):
            publisher.chart_line(record['id'], record['samples'],
                                 record['meta'], x_title='time, s')

        write_record_stats(publisher, record)


def write_stats(publisher, records, row_header, show_all=False):
    if len(records) < 1:
        return

    publisher.subheader('Stats')
    records.sort(key=lambda x: x[row_header])

    if show_all:
        keys = ['min', 'avg', 'max']
    else:
        keys = ['avg']
    meta = []
    headers = []

    # collect meta
    record = records[0]
    headers.append(row_header)

    for param, values in record['stats'].items():
        for key in keys:
            header = ''
            if show_all:
                header = key + ' '
            header += param
            if values['unit']:
                header += ', ' + values['unit']
            headers.append(header)
            meta.append((param, key))

    # fill the table
    table = [headers]
    for record in records:
        row = [record[row_header]]
        for m in meta:
            param, key = m

            if param in record['stats']:
                row.append(record['stats'][param][key])
            else:
                row.append('n/a')

        table.append(row)

    publisher.table(table)


def write_book(doc_folder, data):
    records = data['records'].values()
    publisher = ReSTPublisher(doc_folder)

    for scenario in data['scenarios'].keys():
        publisher.ref_label(scenario)
        publisher.header(scenario)

        scenario_def = data['scenarios'][scenario]
        if 'description' in scenario_def:
            publisher.para(scenario_def['description'])

        write_scenario_definition(publisher, scenario_def)

        write_errors(publisher, filter_records(records, scenario=scenario))

        test_records = filter_records(records, type='test', scenario=scenario)
        test_records.sort(key=lambda x: x['test'])

        for record in test_records:
            test = record['test']
            publisher.header(test, level=1)

            write_test_definition(data, publisher, test)

            concurrency_records = filter_records(records, type='concurrency',
                                                 scenario=scenario, test=test)
            concurrency_records.sort(key=lambda x: int(x['concurrency']))

            concurrency_count = len(concurrency_records)

            if concurrency_count >= 2:
                if record.get('chart'):
                    publisher.chart_xy(record['id'], record['chart'],
                                       record['meta'], 'concurrency')
                    write_stats(publisher, concurrency_records, 'concurrency')

            write_sla(publisher, concurrency_records, data['sla'])

            write_concurrency_block(publisher, records, concurrency_records,
                                    data['sla'])



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import testtools

from shaker.engine.executors import base
from shaker.engine import quorum


class TestOperations(testtools.TestCase):

    def test_execute_operation_process_reply(self):
        executor = mock.MagicMock()
        executor.process_reply = mock.Mock(return_value={'samples': []})

        agent_id = 'the-agent'
        ex = quorum.ExecuteOperation({agent_id: executor})

        message = {
            'stdout': 'foo',
            'stderr': '',
        }
        reply = ex.process_reply(agent_id, message)

        expected = {
            'status': 'ok',
            'samples': [],
        }
        executor.process_reply.assert_called_once_with(message)
        self.assertEqual(expected, reply)

    def test_execute_operation_process_reply_with_error(self):
        executor = mock.MagicMock()
        executor.process_reply = mock.Mock(
            side_effect=base.ExecutorException({'stderr': 'sad'}, 'Error!'))

        agent_id = 'the-agent'
        ex = quorum.ExecuteOperation({agent_id: executor})

        message = {
            'stdout': 'foo',
            'stderr': '',
        }
        reply = ex.process_reply(agent_id, message)

        expected = {
            'status': 'error',
            'stderr': 'sad',
            'info': 'Error!'
        }
        executor.process_reply.assert_called_once_with(message)
        self.assertDictContainsSubset(expected, reply)

    def test_execute_operation_process_reply_with_unhandled_exception(self):
        executor = mock.MagicMock()
        executor.process_reply = mock.Mock(
            side_effect=Exception('Boom!'))

        agent_id = 'the-agent'
        ex = quorum.ExecuteOperation({agent_id: executor})

        message = {}
        reply = ex.process_reply(agent_id, message)

        expected = {
            'status': 'error',
            'info': 'Boom!'
        }
        executor.process_reply.assert_called_once_with(message)
        self.assertDictContainsSubset(expected, reply)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from shaker.engine import config
from shaker.engine import server
from shaker.engine import utils


def main():
    utils.init_config_and_logging(
        config.SCENARIO_OPTS + config.REPORT_OPTS
    )

    server.act()


if __name__ == "__main__":
    main()



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from shaker.engine.aggregators import base


def _filter_none(array):
    return [x for x in array if x is not None]


def avg(array):
    s = _filter_none(array)
    return sum(s) / len(s) if s else 0


def safe_max(array):
    s = _filter_none(array)
    return max(s) if s else None


def safe_min(array):
    s = _filter_none(array)
    return min(s) if s else None


class TrafficAggregator(base.BaseAggregator):
    def __init__(self, test_definition):
        super(TrafficAggregator, self).__init__(test_definition)

    def test_summary(self, records):
        chart = []
        xs = []
        avg_v = collections.defaultdict(list)
        units = {}

        for record in sorted(records, key=lambda x: x['concurrency']):
            xs.append(record['concurrency'])
            for k, v in record['stats'].items():
                avg_v[k].append(v['avg'])
                units[k] = v['unit']

        chart.append(['concurrency'] + xs)
        meta = [('concurrency', '')]

        for k in avg_v.keys():
            chart.append([k] + avg_v[k])
            meta.append((k, units[k]))

        return {
            'chart': chart,
            'meta': meta,
        }

    def concurrency_summary(self, records):
        max_v = collections.defaultdict(list)
        min_v = collections.defaultdict(list)
        avg_v = collections.defaultdict(list)
        unit_v = dict()

        nodes = []
        for record in records:
            nodes.append(record['node'])

            for k, v in record['stats'].items():
                if 'max' in v:
                    max_v[k].append(v['max'])
                if 'min' in v:
                    min_v[k].append(v['min'])
                if 'avg' in v:
                    avg_v[k].append(v['avg'])
                if 'unit' in v:
                    unit_v[k] = v['unit']

        stats = {}
        node_chart = [['x'] + nodes]

        for k in unit_v.keys():
            stats[k] = dict()
            title = k
            if k in unit_v:
                stats[k]['unit'] = unit_v[k]
                title += ', ' + unit_v[k]
            if avg_v[k]:
                stats[k]['avg'] = avg(avg_v[k])
                node_chart.append(['Avg %s' % title] + avg_v[k])
            if max_v[k]:
                stats[k]['max'] = max(max_v[k])
                node_chart.append(['Max %s' % title] + max_v[k])
            if min_v[k]:
                stats[k]['min'] = min(min_v[k])
                node_chart.append(['Min %s' % title] + min_v[k])

        return {
            'stats': stats,
            'node_chart': node_chart,
        }

    def record_summary(self, record):
        # convert bit/s to Mbit/s
        for idx, item_meta in enumerate(record.get('meta', [])):
            if item_meta[1] in ['bit/s', 'bits/s']:
                for row in record.get('samples'):
                    if row[idx]:
                        row[idx] = float(row[idx]) / 1024 / 1024
                item_meta[1] = 'Mbit/s'

        # calculate stats
        record['stats'] = record.get('stats') or dict()

        for idx, item_meta in enumerate(record.get('meta', [])):
            column = [row[idx] for row in record.get('samples')]

            item_title = item_meta[0]
            if item_title != 'time':
                record['stats'][item_title] = {
                    'max': safe_max(column),
                    'min': safe_min(column),
                    'avg': avg(column),
                    'unit': item_meta[1],
                }

        # drop stdout
        if 'stdout' in record:
            del record['stdout']
